--[[
Custom text area like thing for displaying lyrics. Optionally parses
LRC format timeing and automaticall scrolls lyrics in time with the
song!

Key bindings:
 - Scroll wheel: Up/down lyrics
 - Left: Exit to lyrics match list
 - Play: Sync to current line
 - Add: Sync to current elapsed time


--]]

-- stuff we use 

local pairs, ipairs, tostring, type = pairs, ipairs, tostring, type


local oo                     = require("loop.simple")

local string                 = require("string") 
local table                  = require("jive.utils.table") 
local io                     = require("io")

local Applet                 = require("jive.Applet") 
local Framework              = require("jive.ui.Framework") 
local Label                  = require("jive.ui.Label") 
local SimpleMenu             = require("jive.ui.SimpleMenu") 
local Textarea               = require("jive.ui.Textarea") 
local Window                 = require("jive.ui.Window") 
local Popup                  = require("jive.ui.Popup")

local socket           = require("socket") 
local SocketHttp       = require("jive.net.SocketHttp") 
local RequestHttp      = require("jive.net.RequestHttp") 
local appletManager    = appletManager 
local RequestJsonRpc = require("jive.net.RequestJsonRpc") 
local Timer            = require("jive.ui.Timer") 
local Widget = require("jive.ui.Widget") 

local EVENT_ALL      = jive.ui.EVENT_ALL 
local EVENT_KEY_PRESS      = jive.ui.EVENT_KEY_PRESS 
local KEY_LEFT      = jive.ui.KEY_LEFT
local KEY_BACK      = jive.ui.KEY_BACK 
local KEY_PLAY      = jive.ui.KEY_PLAY 
local KEY_ADD = jive.ui.KEY_ADD
local EVENT_SCROLL      = jive.ui.EVENT_SCROLL 
local EVENT_CONSUME    = jive.ui.EVENT_CONSUME
local EVENT_UNUSED     = jive.ui.EVENT_UNUSED


local jnt              = jnt

local log                    = require("jive.utils.log").logger("applets.setup")

module(...) 


-- Represents a single line of a lyric, which may map to one or more
-- physical lines of text on the display
DLine = oo.class()

function DLine:__init(txt, width)
   local obj = oo.rawnew(self)
   local cursor = 0
   obj.lines = {}
   obj.numLines = 0
   local t = {}
   -- Split into words + wrap wholwe words at 'width'
   for w in txt:gmatch("%w+") do
      if #w + cursor > width then
	 -- Start a new line
	 table.insert(obj.lines, Label("lyrics", table.concat(t, ' ')))
	 obj.numLines = obj.numLines + 1
	 cursor = 0
	 t = {}
      end
      cursor = cursor + #w + 1
      table.insert(t, w)
   end
   if #t then
      table.insert(obj.lines, Label("lyrics", table.concat(t, ' ')))
      obj.numLines = obj.numLines + 1
   end
   return obj
end


function DLine:length()
   return self.numLines
end

function DLine:addWidgets(w)
   for _, v in pairs(self.lines) do
      w:addWidget(v)
   end
end

function DLine:removeWidgets(w)
   for _, v in pairs(self.lines) do
      w:removeWidget(v)
   end
end

function DLine:setStyle(style)
   for _, v in pairs(self.lines) do
      v:setStyle(style)
   end
end

-------------------------------------------

-- A scrollable text area that:
--  Wraps lines at word boundaries
--  Scrolls wrapped lines as a single entity
--  Highlights current line
--  Parses LRC and automatically scrolls text in time

MyTextArea = oo.class()

function MyTextArea:__init(text, nk, w, getTimer) 
   log:debug("MyTextArea:__init") 
   local obj = oo.rawnew(self) 
   obj.w = w
   obj.getTimer = getTimer
   obj.alive = true
   obj.currLine = 1
   obj.visibleLength = 12  -- Lines
   obj.visibleWidth = 32  -- Chars
   obj.dlines = {} 
   obj.times = {}
   obj.numWidgets = 0
   text = text .. '\n'
   for l in text:gmatch("(.-)[\r\n]+") do 
      log:debug("Add: " .. l)
      if not nk then
	 local ts = l:match('%[(.*)%]')
	 if ts then
	    i, t = self:parseTime(ts)
	    l = l:sub(i)
	    if l:match("%w") then
	       table.insert(obj.times, t)
	    end
	 else
	    obj.times = {}
	    nk = false -- Give up with karaoke
	 end
      end
      if l:match("%w") then
	 local dl = DLine(l, obj.visibleWidth)
	 obj.numWidgets = obj.numWidgets + 1
	 table.insert(obj.dlines, dl)
      end
   end
   obj.dlines[1]:setStyle("current")
   if #obj.times ~= 0 then
      obj:syncToCurrentTime()
   end
   return obj 
end

-- Scroll to the next lyric and set a timer for the next scroll
function MyTextArea:nextLyricTime()
   if self.currLine < self.numWidgets and self.alive then
      local nextIn =  self.times[self.currLine + 1] - self.times[self.currLine]
      log:debug("Scroll to line: " .. self.currLine + 1 .. " for " .. nextIn .. "mS")
      self.timer = Timer(nextIn, function() MyTextArea.nextLyricTime(self) end, true)
      self.timer:start()
      self:scrollBy(1, self.w)
   end
end

-- Synchronise automatic scrolling to current line
function MyTextArea:syncToCurrentLine(adjustTo)
   log:debug("syncToCurrentLine")
   if #self.times ~= 0 then
      if self.timer then
	 self.timer:stop()
      end
      local nextIn =  self.times[self.currLine + 1] - self.times[self.currLine]
      if adjustTo then
	 nextIn = self.times[self.currLine + 1] - adjustTo
      end
      log:debug("Sync to line: " .. self.currLine .. " for " .. nextIn .. "mS")
      self.timer = Timer(nextIn, function() MyTextArea.nextLyricTime(self) end, true)
      self.timer:start()
   end
end

-- Synchronize display to elapsed time for current song, iff local
function MyTextArea:syncToCurrentTime()
   log:debug("syncToCurrentTime")
   self.getTimer(function(currTime, remote)
		    log:debug('currTime: ' .. currTime)
		    if currTime != 0 or remote then
		       -- Not paused/stopped
		       currTime = currTime * 1000
		       for i, v in pairs(self.times) do
			  log:debug(string.format("i: %d, v: %f", i, v))
			  if v >= currTime then
			     if self.timer then
				self.timer:stop()
			     end
			     -- Backup 1 line
			     if i > 1 then
				i = i - 1
			     end
			     self:scrollBy(i - self.currLine, self.w)
			     self:syncToCurrentLine(currTime)
			     return
			  end
		       end
		    end
		 end)
end

function MyTextArea:parseTime(ts)
   local mins, secs, hsecs
   mins, secs, hsecs = ts:match("(.*):(.*)%.*(.*)")
   x = 60000 * mins + secs * 1000
   -- 1/100ths seconds not always present
   if hsecs and hsecs ~= "" then
      x = x + hsecs * 10
   end
   return ts:len() + 2, x
end

function MyTextArea:scrollBy(scroll, w) 
   log:debug("scrollBy " .. scroll)
   -- Turn off highlight for current line
   self.dlines[self.currLine]:setStyle("lyrics")

   self.currLine = self.currLine + scroll
   if self.currLine < 1 then
      self.currLine = 1
   end
   if self.currLine >= self.numWidgets then
      self.currLine = self.numWidgets 
   end

   self.dlines[self.currLine]:setStyle("current")
   self:draw(w) 
end

function MyTextArea:draw() 
   for _, v in pairs(self.dlines) do
      v:removeWidgets(self.w)
   end

   local dline = 1
   local widgets = {}
   local startLine = 1

   -- Display current line with some context by working out
   -- how many real lines of display are used.
   local clr = 0
   for i = 1, self.currLine do
      clr = clr + self.dlines[i]:length()
   end
	     
   if clr >= self.visibleLength/2 then
      startLine = self.currLine - 2
   end
   
   for i = startLine, self.numWidgets do
      local l = self.dlines[i]:length()
      -- Can these labels fit into the display?
      if dline + l > self.visibleLength then
	 break -- Screen full
      end
      -- Can fit
      self.dlines[i]:addWidgets(self.w)
      dline = dline + l
   end 
end

function MyTextArea:setEventHandler()
   self.w:addListener(EVENT_ALL,
		      function (event)
			 return self:_eventHandler(event)
		      end)

end

function MyTextArea:_eventHandler(event)
   local type = event:getType()
   if type == EVENT_SCROLL then
      if self.timer then
	 self.timer:stop()
      end
      self:scrollBy(event:getScroll(), self.w)
      return EVENT_CONSUME
   elseif type == EVENT_KEY_PRESS then
      local keycode = event:getKeycode()
      if keycode == KEY_BACK or keycode == KEY_LEFT then
	 -- Go away!
	 self.alive = nil
	 if self.timer then
	    self.timer:stop()
	 end
	 self.w:showInstead()
	 return EVENT_CONSUME
      elseif keycode == KEY_PLAY then
	 self:syncToCurrentLine()
      elseif keycode == KEY_ADD then
	 self:syncToCurrentTime()
      end
   end
   return EVENT_UNUSED
end

